"""4x4 matrices for rotations and translations.

Functions in this module generate matrices which left-multiply column vectors containing
`xyzw` or `zyxw` homogenous coordinates.
"""

import torch
import einops


def Rx(angles: torch.Tensor | list | tuple | float, 
       zyx: bool = False, 
       device: torch.device | None = None) -> torch.Tensor:
    """4x4 matrices for a rotation of homogenous coordinates around the X-axis.

    Matrix structure (xyzw):
    ┌             ┐
    │ 1  0   0  0 │
    │ 0  c  -s  0 │
    │ 0  s   c  0 │
    │ 0  0   0  1 │
    └             ┘

    Matrix structure (zyxw):
    ┌             ┐
    │ 1  0   0  0 │
    │ 0  c   s  0 │
    │ 0 -s   c  0 │
    │ 0  0   0  1 │
    └             ┘
    where c=cos(θ), s=sin(θ)

    Parameters
    ----------
    angles: torch.Tensor | list | tuple | float
        `(..., )` array, list-like, or single float of angles in degrees.
    zyx: bool
        Whether output should be compatible with `zyxw` (`True`) or `xyzw`
        (`False`) homogenous coordinates.
    device: torch.device, optional
        The device on which to place the resulting tensor. If None, uses the
        device of the input tensor or defaults to CPU.

    Returns
    -------
    matrices: `(..., 4, 4)` array of 4x4 rotation matrices.
    """
    # shape (...) -> (n, )
    angles = torch.as_tensor(angles, dtype=torch.float32)
    device = device or angles.device  # Use provided device or input tensor's device
    angles = angles.to(device)
    angles_packed, ps = einops.pack([angles], pattern='*')  # to 1d
    n = angles_packed.shape[0]

    # calculate useful values
    angles_radians = torch.deg2rad(angles_packed)
    c = torch.cos(angles_radians)
    s = torch.sin(angles_radians)

    # construct matrices
    matrices = einops.repeat(torch.eye(4, device=device), 'i j -> n i j', n=n).clone()
    matrices[:, 1, 1] = c
    matrices[:, 1, 2] = -s
    matrices[:, 2, 1] = s
    matrices[:, 2, 2] = c

    # operating on zyx coordinates?
    if zyx is True:
        matrices[:, :3, :3] = torch.flip(matrices[:, :3, :3], dims=(-2, -1))

    # shape (n, 4, 4) -> (..., 4, 4)
    [matrices] = einops.unpack(matrices, packed_shapes=ps, pattern='* i j')
    return matrices


def Ry(angles: torch.Tensor | list | tuple | float, 
       zyx: bool = False, 
       device: torch.device | None = None) -> torch.Tensor:
    """4x4 matrices for a rotation of homogenous coordinates around the Y-axis.

    Matrix structure (xyzw):
    ┌             ┐
    │ c  0  s  0  │
    │ 0  1  0  0  │
    │-s  0  c  0  │
    │ 0  0  0  1  │
    └             ┘

    Matrix structure (zyxw):
    ┌             ┐
    │ c  0 -s  0  │
    │ 0  1  0  0  │
    │ s  0  c  0  │
    │ 0  0  0  1  │
    └             ┘
    where c=cos(θ), s=sin(θ)

    Parameters
    ----------
    angles: torch.Tensor | list | tuple | float
        `(..., )` array, list-like, or single float of angles in degrees.
    zyx: bool
        Whether output should be compatible with `zyxw` (`True`) or `xyzw`
        (`False`) homogenous coordinates.
    device: torch.device, optional
        The device on which to place the resulting tensor. If None, uses the
        device of the input tensor or defaults to CPU.

    Returns
    -------
    matrices: `(..., 4, 4)` array of 4x4 rotation matrices.
    """
    # shape (...) -> (n, )
    angles = torch.as_tensor(angles, dtype=torch.float32)
    device = device or angles.device  # Use provided device or input tensor's device
    angles = angles.to(device)
    angles_packed, ps = einops.pack([angles], pattern='*')  # to 1d
    n = angles_packed.shape[0]

    # calculate useful values
    angles_radians = torch.deg2rad(angles_packed)
    c = torch.cos(angles_radians)
    s = torch.sin(angles_radians)

    # construct matrices
    matrices = einops.repeat(torch.eye(4, device=device), 'i j -> n i j', n=n).clone()
    matrices[:, 0, 0] = c
    matrices[:, 0, 2] = s
    matrices[:, 2, 0] = -s
    matrices[:, 2, 2] = c

    # operating on zyx coordinates?
    if zyx is True:
        matrices[:, :3, :3] = torch.flip(matrices[:, :3, :3], dims=(-2, -1))

    # shape (n, 4, 4) -> (..., 4, 4)
    [matrices] = einops.unpack(matrices, packed_shapes=ps, pattern='* i j')
    return matrices


def Rz(angles: torch.Tensor | list | tuple | float, 
       zyx: bool = False, 
       device: torch.device | None = None) -> torch.Tensor:
    """4x4 matrices for a rotation of homogenous coordinates around the Z-axis.

    Matrix structure (xyzw):
    ┌             ┐
    │ c -s  0  0  │
    │ s  c  0  0  │
    │ 0  0  1  0  │
    │ 0  0  0  1  │
    └             ┘

    Matrix structure (zyxw):
    ┌             ┐
    │ c  s  0  0  │
    │-s  c  0  0  │
    │ 0  0  1  0  │
    │ 0  0  0  1  │
    └             ┘
    where c=cos(θ), s=sin(θ)

    Parameters
    ----------
    angles: torch.Tensor | list | tuple | float
        `(..., )` array, list-like, or single float of angles in degrees.
    zyx: bool
        Whether output should be compatible with `zyxw` (`True`) or `xyzw`
        (`False`) homogenous coordinates.
    device: torch.device, optional
        The device on which to place the resulting tensor. If None, uses the
        device of the input tensor or defaults to CPU.

    Returns
    -------
    matrices: `(..., 4, 4)` array of 4x4 rotation matrices.
    """
    # shape (...) -> (n, )
    angles = torch.as_tensor(angles, dtype=torch.float32)
    device = device or angles.device  # Use provided device or input tensor's device
    angles = angles.to(device)
    angles_packed, ps = einops.pack([angles], pattern='*')  # to 1d
    n = angles_packed.shape[0]

    # calculate useful values
    angles_radians = torch.deg2rad(angles_packed)
    c = torch.cos(angles_radians)
    s = torch.sin(angles_radians)
    matrices = einops.repeat(torch.eye(4, device=device), 'i j -> n i j', n=n).clone()
    matrices[:, 0, 0] = c
    matrices[:, 0, 1] = -s
    matrices[:, 1, 0] = s
    matrices[:, 1, 1] = c
    if zyx is True:
        matrices[:, :3, :3] = torch.flip(matrices[:, :3, :3], dims=(-2, -1))

    # shape (n, 4, 4) -> (..., 4, 4)
    [matrices] = einops.unpack(matrices, packed_shapes=ps, pattern='* i j')
    return matrices


def T(shifts: torch.Tensor | list | tuple, 
      device: torch.device | None = None) -> torch.Tensor:
    """4x4 matrices for translations.

    Matrix structure:
    ┌             ┐
    │ 1  0  0  t0 │
    │ 0  1  0  t1 │
    │ 0  0  1  t2 │
    │ 0  0  0  1  │
    └             ┘
    where t0, t1, t2 are translation components at indices 0, 1, 2 in the last
    dimension of `shifts`.

    Parameters
    ----------
    shifts: torch.Tensor | list | tuple
        `(..., 3)` array of shifts.
    device: torch.device, optional
        The device on which to place the resulting tensor. If None, uses the
        device of the input tensor or defaults to CPU.

    Returns
    -------
    matrices: torch.Tensor
        `(..., 4, 4)` array of 4x4 shift matrices.
    """
    # shape (...) -> (n, )
    shifts = torch.as_tensor(shifts, dtype=torch.float32)
    if shifts.ndim > 0 and shifts.shape[-1] != 3:
        raise ValueError("Shifts must have the last dimension of size 3 for 3D transformations.")
    shifts = torch.atleast_1d(shifts)
    device = device or shifts.device  # Use provided device or input tensor's device
    shifts = shifts.to(device)
    shifts, ps = einops.pack([shifts], pattern='* coords')  # to 2d
    n = shifts.shape[0]

    # construct matrices
    matrices = einops.repeat(torch.eye(4, device=device), 'i j -> n i j', n=n).clone()
    matrices[:, :3, 3] = shifts

    # shape (n, 4, 4) -> (..., 4, 4)
    [matrices] = einops.unpack(matrices, packed_shapes=ps, pattern='* i j')
    return matrices


def S(scale_factors: torch.Tensor | list | tuple | float, 
      device: torch.device | None = None) -> torch.Tensor:
    """4x4 matrices for scaling.

     Matrix structure:
    ┌             ┐
    │ s1 0  0  0  │
    │ 0  s2 0  0  │
    │ 0  0  s3 0  │
    │ 0  0  0  1  │
    └             ┘
    where s1, s2, s3 are the scaling factors at indices 0, 1, 2 in the last
    dimension of `scale_factors`.

    Parameters
    ----------
    scale_factors: torch.Tensor | list | tuple | float
        `(..., 3)` array, list-like, or single float of scale factors.
    device: torch.device, optional
        The device on which to place the resulting tensor. If None, uses the
        device of the input tensor or defaults to CPU.

    Returns
    -------
    matrices: torch.Tensor
        `(..., 4, 4)` array of 4x4 shift matrices.
    """
    # shape (...) -> (n, )
    scale_factors = torch.as_tensor(scale_factors, dtype=torch.float32)
    if scale_factors.ndim > 0 and scale_factors.shape[-1] != 3:
        raise ValueError("Scale factors must have the last dimension of size 3 for 3D transformations.")
    scale_factors = torch.atleast_1d(scale_factors)
    device = device or scale_factors.device  # Use provided device or input tensor's device
    scale_factors = scale_factors.to(device)
    scale_factors, ps = einops.pack([scale_factors], pattern='* coords')  # to 2d
    n = scale_factors.shape[0]

    # construct matrices
    matrices = einops.repeat(torch.eye(4, device=device), 'i j -> n i j', n=n).clone()
    matrices[:, [0, 1, 2], [0, 1, 2]] = scale_factors

    # shape (n, 4, 4) -> (..., 4, 4)
    [matrices] = einops.unpack(matrices, packed_shapes=ps, pattern='* i j')
    return matrices



"""Utilities for affine transformations of 2d/3d coordinates in PyTorch"""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("torch-affine-utils")
except PackageNotFoundError:
    __version__ = "uninstalled"

__author__ = "Alister Burt"
__email__ = "alisterburt@gmail.com"

from torch_affine_utils.utils import homogenise_coordinates

__all__ = [
    "homogenise_coordinates",
]


"""3x3 matrices for rotations and translations in 2D.

Functions in this module generate matrices which left-multiply column vectors containing
`xyw` or `yxw` homogenous coordinates.
"""

import torch
import einops


def R(angles: torch.Tensor | list | tuple | float, 
      yx: bool = False, 
      device: torch.device | None = None) -> torch.Tensor:
    """3x3 matrices for a rotation of homogenous coordinates in 2D.

    Matrix structure (xyw):
    ┌          ┐
    │ c -s  0  │
    │ s  c  0  │
    │ 0  0  1  │
    └          ┘

    Matrix structure (yxw):
    ┌          ┐
    │ c  s  0  │
    │-s  c  0  │
    │ 0  0  1  │
    └          ┘
    where c=cos(θ), s=sin(θ)

    Parameters
    ----------
    angles: torch.Tensor | list | tuple | float
        `(..., )` array, list-like, or single float of angles in degrees.
    yx: bool
        Whether output should be compatible with `yxw` (`True`) or `xyw`
        (`False`) homogenous coordinates.
    device: torch.device, optional
        The device on which to place the resulting tensor. If None, uses the
        device of the input tensor or defaults to CPU.

    Returns
    -------
    matrices: `(..., 3, 3)` array of 3x3 rotation matrices.
    """
    angles = torch.as_tensor(angles, dtype=torch.float32)
    device = device or angles.device  # Use provided device or input tensor's device
    angles = angles.to(device)
    angles_packed, ps = einops.pack([angles], pattern='*')  # to 1d
    n = angles_packed.shape[0]

    # calculate useful values
    angles_radians = torch.deg2rad(angles_packed)
    c = torch.cos(angles_radians)
    s = torch.sin(angles_radians)

    # construct matrices
    matrices = einops.repeat(torch.eye(3, device=device), 'i j -> n i j', n=n).clone()
    matrices[:, 0, 0] = c
    matrices[:, 0, 1] = -s
    matrices[:, 1, 0] = s
    matrices[:, 1, 1] = c

    # operating on yx coordinates?
    if yx is True:
        matrices[:, :2, :2] = torch.flip(matrices[:, :2, :2], dims=(-2, -1))

    # shape (n, 3, 3) -> (..., 3, 3)
    [matrices] = einops.unpack(matrices, packed_shapes=ps, pattern='* i j')
    return matrices


def T(shifts: torch.Tensor | list | tuple, 
      device: torch.device = None) -> torch.Tensor:
    """3x3 matrices for translations in 2D.

    Matrix structure:
    ┌           ┐
    │ 1  0  t0  │
    │ 0  1  t1  │
    │ 0  0  1   │
    └           ┘
    where t0, t1 are translation components at indices 0, 1 in the last
    dimension of `shifts`.

    Parameters
    ----------
    shifts: torch.Tensor | list | tuple
        `(..., 2)` array or list-like of shifts.
    device: torch.device, optional
        The device on which to place the resulting tensor. If None, uses the
        device of the input tensor or defaults to CPU.

    Returns
    -------
    matrices: torch.Tensor
        `(..., 3, 3)` array of 3x3 shift matrices.
    """
    shifts = torch.as_tensor(shifts, dtype=torch.float32)
    if shifts.ndim > 0 and shifts.shape[-1] != 2:
        raise ValueError("Shifts must have the last dimension of size 2 for 2D transformations.")
    shifts = torch.atleast_1d(shifts)
    device = device or shifts.device  # Use provided device or input tensor's device
    shifts = shifts.to(device)
    shifts, ps = einops.pack([shifts], pattern='* coords')  # to 2d
    n = shifts.shape[0]

    # construct matrices
    matrices = einops.repeat(torch.eye(3, device=device), 'i j -> n i j', n=n).clone()
    matrices[:, :2, 2] = shifts

    # shape (n, 3, 3) -> (..., 3, 3)
    [matrices] = einops.unpack(matrices, packed_shapes=ps, pattern='* i j')
    return matrices


def S(scale_factors: torch.Tensor | list | tuple | float, 
      device: torch.device | None = None) -> torch.Tensor:
    """3x3 matrices for scaling in 2D.

     Matrix structure:
    ┌          ┐
    │ s1 0  0  │
    │ 0  s2 0  │
    │ 0  0  1  │
    └          ┘
    where s1, s2 are the scaling factors at indices 0, 1 in the last dimension
    of `scale_factors`.

    Parameters
    ----------
    scale_factors: torch.Tensor | list | tuple | float
        `(..., 2)` array, list-like, or single float of scale factors.
    device: torch.device, optional
        The device on which to place the resulting tensor. If None, uses the
        device of the input tensor or defaults to CPU.

    Returns
    -------
    matrices: torch.Tensor
        `(..., 3, 3)` array of 3x3 scaling matrices.
    """
    scale_factors = torch.as_tensor(scale_factors, dtype=torch.float32)
    if scale_factors.ndim > 0 and scale_factors.shape[-1] != 2:
        raise ValueError("Scale factors must have the last dimension of size 2 for 2D transformations.")
    scale_factors = torch.atleast_1d(scale_factors)
    device = device or scale_factors.device  # Use provided device or input tensor's device
    scale_factors = scale_factors.to(device)
    scale_factors, ps = einops.pack([scale_factors], pattern='* coords')  # to 2d
    n = scale_factors.shape[0]

    # construct matrices
    matrices = einops.repeat(torch.eye(3, device=device), 'i j -> n i j', n=n).clone()
    matrices[:, [0, 1], [0, 1]] = scale_factors

    # shape (n, 3, 3) -> (..., 3, 3)
    [matrices] = einops.unpack(matrices, packed_shapes=ps, pattern='* i j')
    return matrices



import torch

from torch_affine_utils.transforms_3d import Rx, Ry, Rz, T, S

TRANSFORMS = [Rx, Ry, Rz, T, S]

def test_rotation_around_x():
    """Rotation of y around x should become z."""
    R = Rx(90)
    v = torch.tensor([0, 1, 0, 1]).view((4, 1)).float()
    expected = torch.tensor([0, 0, 1, 1]).view((4, 1)).float()
    assert torch.allclose(R @ v, expected, atol=1e-6)

    R = Rx(90, zyx=True)
    v = torch.tensor([0, 1, 0, 1]).view((4, 1)).float()
    expected = torch.tensor([1, 0, 0, 1]).view((4, 1)).float()
    assert torch.allclose(R @ v, expected, atol=1e-6)


def test_rotation_around_y():
    """Rotation of z around y should be x"""
    R = Ry(90)
    v = torch.tensor([0, 0, 1, 1]).view((4, 1)).float()
    expected = torch.tensor([1, 0, 0, 1]).view((4, 1)).float()
    assert torch.allclose(R @ v, expected, atol=1e-6)

    R = Ry(90, zyx=True)
    v = torch.tensor([1, 0, 0, 1]).view((4, 1)).float()
    expected = torch.tensor([0, 0, 1, 1]).view((4, 1)).float()
    assert torch.allclose(R @ v, expected, atol=1e-6)


def test_rotation_around_z():
    """Rotation of x around z should give y."""
    R = Rz(90)
    v = torch.tensor([1, 0, 0, 1]).view((4, 1)).float()
    expected = torch.tensor([0, 1, 0, 1]).view((4, 1)).float()
    assert torch.allclose(R @ v, expected, atol=1e-6)

    R = Rz(90, zyx=True)
    v = torch.tensor([0, 0, 1, 1]).view((4, 1)).float()
    expected = torch.tensor([0, 1, 0, 1]).view((4, 1)).float()
    assert torch.allclose(R @ v, expected, atol=1e-6)


def test_translation():
    """Translations"""
    M = T([1, 2, 3])
    v = torch.tensor([0, 0, 0, 1]).view((4, 1)).float()
    expected = torch.tensor([1, 2, 3, 1]).view((4, 1)).float()
    assert torch.allclose(M @ v, expected, atol=1e-6)


def test_scaling():
    """Translations"""
    M = S([1, 2, 3])
    v = torch.tensor([1, 1, 1, 1]).view((4, 1)).float()
    expected = torch.tensor([1, 2, 3, 1]).view((4, 1)).float()
    assert torch.allclose(M @ v, expected, atol=1e-6)


def test_devices():
    """Test that the matrices are created on the correct device."""
    for O in TRANSFORMS:
        assert O(torch.tensor(0)).device.type == 'cpu'
        assert O(torch.tensor(0, device="meta")).device.type == 'meta'
        # Whish we could test this, but it doesn't work
        # because we can't grab data from a meta tensor
        #assert O(torch.tensor(0, device="meta"), device="cpu").device.type == 'cpu'
        assert O(torch.tensor(0), device="meta").device.type == 'meta'


def test_batching():
    """Test that the matrices are created with the correct batch size."""

    # Single values should give 4x4 matrices for all transforms
    for O in TRANSFORMS:
        # Test with a single value
        matrix = O(0)
        assert matrix.shape == (4, 4)

    # 1D tensor should give (n,4,4) matrices for all transforms
    for O in [Rx, Ry, Rz]:
        assert O(torch.tensor([0, 90, 180, 250])).shape == (4, 4, 4)

    # 1D tensor with length 3 should result in 4x4 matrices for T and S
    for O in [T, S]:
        matrix = O(torch.tensor([0, 90, 180]))
        assert matrix.shape == (4, 4)
    
    # 1D tensors with length greater than 3 should raise an error for T and S
    for O in [T, S]:
        try:
            O(torch.tensor([0, 90, 180, 250]))
        except ValueError:
            pass
        else:
            raise AssertionError(f"{O.__name__} should raise an error for 1D tensors with length > 3")

    # 2D tensor should give (m,n,4,4) matrices for all rotations
    for O in [Rx, Ry, Rz]:
        assert O(torch.tensor([[0, 90], [180, 250]])).shape == (2, 2, 4, 4)

    # (n,3) tensor should result in (n,4,4) matrices for T and S
    for O in [T, S]:
        matrix = O(torch.tensor([[0, 0, 0], [5, 5, 5]]))
        assert matrix.shape == (2, 4, 4)

    # (n,>3) tensor should raise an error for T and S
    for O in [T, S]:
        try:
            O(torch.tensor([[0, 0, 0, 0], [5, 5, 5, 5]]))
        except ValueError:
            pass
        else:
            raise AssertionError(f"{O.__name__} should raise an error for tensors with last dimension > 3")


def test_backpropagation():
    """Test that gradients can be back propagated from output to input."""
    for O in TRANSFORMS:
        x = torch.tensor([90.0, 60.0, 30.0], requires_grad=True)
        y = O(x)
        assert y.requires_grad

        # y needs to be a scalar value (i.e. a loss) for backpropagation to work
        # hence the sum() operation
        y.sum().backward()
        assert x.grad is not None


import torch

from torch_affine_utils.transforms_2d import R, T, S

TRANSFORMS = [R, T, S]

def test_rotation():
    """Rotating x-axis by 90 degrees should become y-axis."""
    # Standard coordinate system (xyw)
    rotation = R(90)
    v = torch.tensor([1, 0, 1]).view((3, 1)).float()
    expected = torch.tensor([0, 1, 1]).view((3, 1)).float()
    assert torch.allclose(rotation @ v, expected, atol=1e-6)

    # Test with yx coordinate system
    rotation = R(90, yx=True)
    v = torch.tensor([0, 1, 1]).view((3, 1)).float()
    expected = torch.tensor([1, 0, 1]).view((3, 1)).float()
    assert torch.allclose(rotation @ v, expected, atol=1e-6)


def test_translation():
    """Translation should add the offset to the position."""
    translation = T([2, 3])
    v = torch.tensor([1, 1, 1]).view((3, 1)).float()
    expected = torch.tensor([3, 4, 1]).view((3, 1)).float()
    assert torch.allclose(translation @ v, expected, atol=1e-6)


def test_scaling():
    """Scaling should multiply each coordinate by the scale factor."""
    scaling = S([2, 3])
    v = torch.tensor([1, 1, 1]).view((3, 1)).float()
    expected = torch.tensor([2, 3, 1]).view((3, 1)).float()
    assert torch.allclose(scaling @ v, expected, atol=1e-6)


def test_devices():
    """Test that the matrices are created on the correct device."""
    for O in TRANSFORMS:
        assert O(torch.tensor(0)).device.type == 'cpu'
        assert O(torch.tensor(0,device="meta")).device.type == 'meta'
        # Whish we could test this, but it doesn't work
        # because we can't grab data from a meta tensor
        #assert O(torch.tensor(0,device="meta"),device="cpu").device.type == 'cpu'
        assert O(torch.tensor(0),device="meta").device.type == 'meta'


def test_batching():
    """Test that the matrices are created with the correct batch size."""

    # Single values should give 3x3 matrices for all transforms
    for O in TRANSFORMS:
        # Test with a single value
        matrix = O(0)
        assert matrix.shape == (3, 3)

    # 1D tensor should give (n,3,3) matrices for all transforms
    assert R(torch.tensor([0, 90, 180, 250])).shape == (4, 3, 3)

    # 1D tensor with length 2 should result in 3x3 matrices for T and S
    for O in [T, S]:
        matrix = O(torch.tensor([0, 90]))
        assert matrix.shape == (3, 3)
    
    # 1D tensors with length greater than 2 should raise an error for T and S
    for O in [T, S]:
        try:
            O(torch.tensor([0, 90, 180]))
        except ValueError:
            pass
        else:
            raise AssertionError(f"{O.__name__} should raise an error for 1D tensors with length > 2")

    # 2D tensor should give (m,n,3,3) matrices for all rotations
    assert R(torch.tensor([[0, 90], [180, 250]])).shape == (2, 2, 3, 3)

    # (n,2) tensor should result in (n,3,3) matrices for T and S
    for O in [T, S]:
        matrix = O(torch.tensor([[0, 0], [5, 5]]))
        assert matrix.shape == (2, 3, 3)

    # (n,>2) tensor should raise an error for T and S
    for O in [T, S]:
        try:
            O(torch.tensor([[0, 0, 0], [5, 5, 5]]))
        except ValueError:
            pass
        else:
            raise AssertionError(f"{O.__name__} should raise an error for tensors with last dimension > 2")


def test_backpropagation_gradients():
    """Test that gradients can be back propagated from output to input."""
    for O in TRANSFORMS:
        x = torch.tensor([90.0, 45.0], requires_grad=True)
        y = O(x)
        assert y.requires_grad

        # y needs to be a scalar value (i.e. a loss) for backpropagation to work
        # hence the sum() operation
        y.sum().backward()
        assert x.grad is not None


import torch
import torch.nn.functional as F


def homogenise_coordinates(coords: torch.Tensor) -> torch.Tensor:
    """Homogenous coordinates to normal coordinates.

    Parameters
    ----------
    coords: torch.Tensor
        `(..., d)` array of d-dimensional coordinates

    Returns
    -------
    output: torch.Tensor
        `(..., d+1)` array of homogenous coordinates
    """
    return F.pad(torch.as_tensor(coords), pad=(0, 1), mode='constant', value=1)
